{*************************************************************************
                Copyright (c) PilotLogic Software House

  Package pl_MagicScript
  This file is part of CodeTyphon Studio (https://www.pilotlogic.com/)

 ***** BEGIN LICENSE BLOCK *****
 * Version: LGPL-2.1 with link exception (Modified LGPL)
 *
 * The contents of this file are subject to the
 * GNU LIBRARY GENERAL PUBLIC LICENSE Version 2.1 (the "License")
 * with the following modification:
 *
 * As a special exception, the copyright holders of this library give you
 * permission to link this library with independent modules to produce an
 * executable, regardless of the license terms of these independent modules,
 * and to copy and distribute the resulting executable under terms of your choice,
 * provided that you also meet, for each linked independent module, the terms
 * and conditions of the license of that module. An independent module is a
 * module which is not derived from or based on this library. If you modify this
 * library, you may extend this exception to your version of the library, but
 * you are not obligated to do so. If you do not wish to do so, delete this
 * exception statement from your version.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * ***** END LICENSE BLOCK *****
 
***************************************************************************}

Unit msbaseobjects;

Interface

{$I magicscript.inc}

Uses
   SysUtils, Classes, TypInfo;

Type

   TmscrIdentifierCharset = Set Of ansichar;

   TmscrDATAItem = Class(TObject)
   Private
      FData: Pointer;
      FItems: TList;
      FName: String;
      FParent: TmscrDATAItem;
      FText: String;
      Function GetCount: integer;
      Function GetItems(Index: integer): TmscrDATAItem;
      Function GetProp(Index: String): String;
      Procedure SetProp(Index: String; Const Value: String);
      Procedure SetParent(Const Value: TmscrDATAItem);
   Public
      Destructor Destroy; Override;
      Procedure Clear;
      Function Root: TmscrDATAItem;
      Function Add: TmscrDATAItem;
      Procedure AddItem(Item: TmscrDATAItem);
      Procedure Assign(Item: TmscrDATAItem);
      Procedure InsertItem(Index: integer; Item: TmscrDATAItem);
      Function Find(Const Name: String): integer;
      Function FindItem(Const Name: String): TmscrDATAItem;
      Function IndexOf(Item: TmscrDATAItem): integer;
      Function PropExists(Const Index: String): boolean;
      Property Data: Pointer Read FData Write FData;
      Property Name: String Read FName Write FName;
      Property Text: String Read FText Write FText;
      Property Count: integer Read GetCount;
      Property Items[Index: integer]: TmscrDATAItem Read GetItems; Default;
      Property Parent: TmscrDATAItem Read FParent Write SetParent;
      Property Prop[Index: String]: String Read GetProp Write SetProp;
   End;

   TmscrDATADocument = Class(TObject)
   Private
      FRoot: TmscrDATAItem;
      FAutoIndent: boolean;
   Public
      Constructor Create;
      Destructor Destroy; Override;
      Procedure Clear;
      Procedure SaveToFile(Const FileName: String);
      Procedure LoadFromFile(Const FileName: String);
      Procedure SaveToStream(Stream: TStream);
      Procedure LoadFromStream(Stream: TStream);
      Property Root: TmscrDATAItem Read FRoot;
      Property AutoIndent: boolean Read FAutoIndent Write FAutoIndent;
   End;

   TmscrDATAReader = Class(TObject)
   Private
      FStream: TStream;
      FSize: int64;
      FPosition: int64;
      FBuffer: pansichar;
      FBufPos: integer;
      FBufEnd: integer;
      Procedure ReadBuffer;
      Procedure ReadItem(Var Name: String; Var Text: String);
      Procedure SetPosition(Const Value: int64);
   Public
      Constructor Create(Stream: TStream);
      Destructor Destroy; Override;
      Procedure ReadHeader;
      Procedure ReadRootItem(Item: TmscrDATAItem);
      Procedure RaiseException;
      Property Size: int64 Read FSize;
      Property Position: int64 Read FPosition Write SetPosition;
   End;

   TmscrDATAWriter = Class(TObject)
   Private
      FStream: TStream;
      FBuffer: ansistring;
      FTempStream: TStream;
      FAutoIndent: boolean;
      Procedure FlushBuffer;
      Procedure WriteItem(Item: TmscrDATAItem; Level: integer = 0);
      Procedure WriteLn(Const s: ansistring);
   Public
      Constructor Create(Stream: TStream);
      Procedure WriteHeader;
      Procedure WriteRootItem(RootItem: TmscrDATAItem);
      Property TempStream: TStream Read FTempStream Write FTempStream;
   End;

   TmscrParser = Class(TObject)
   Private
      FYList: TList;
      FKeywords: TStrings;
      FCaseSensitive: boolean;
      FStringQuotes: String;
      FText: String;
      FCommentBlock1: String;
      FCommentBlock11: String;
      FCommentBlock12: String;
      FCommentBlock2: String;
      FCommentBlock21: String;
      FCommentBlock22: String;
      FCommentLine1: String;
      FCommentLine2: String;
      FHexSequence: String;
      FIdentifierCharset: TmscrIdentifierCharset;
      FSize: integer;
      FPosition: integer;
      FLastPosition: integer;
      FSkipEOL: boolean;
      FSkipSpace: boolean;
      FSpecStrChar: boolean;
      FSkiPChar: String;
      FUseY: boolean;
      Function Ident: String;
      Procedure SetCommentBlock1(Const Value: String);
      Procedure SetCommentBlock2(Const Value: String);
      Function DoDigitSequence: boolean;
      Function DoHexDigitSequence: boolean;
      Function DoScaleFactor: boolean;
      Function DoUnsignedInteger: boolean;
      Function DoUnsignedReal: boolean;
      Procedure SetPosition(Const Value: integer);
      Procedure SetText(Const Value: String);
   Public
      Constructor Create;
      Destructor Destroy; Override;
      Procedure Clear;
      Procedure InitKeywords;
      Procedure ConstructCharset(Const s: String);
      Procedure SkipSpaces;
      Function GetEOL: boolean;
      Function GetIdent: String;
      Function GetChar: String;
      Function GetWord: String;
      Function GetNumber: String;
      Function GetString: String;
      Function GetFRString: String;
      Function GetXYPosition: String;
      Function GetPlainPosition(pt: TPoint): integer;
      Function IsKeyWord(Const s: String): boolean;
      Property Text: String Read FText Write SetText;
      Property Position: integer Read FPosition Write SetPosition;
      Property UseY: boolean Read FUseY Write FUseY;
      Property Keywords: TStrings Read FKeywords;
      Property CaseSensitive: boolean Read FCaseSensitive Write FCaseSensitive;
      Property CommentBlock1: String Read FCommentBlock1 Write SetCommentBlock1;
      Property CommentLine1: String Read FCommentLine1 Write FCommentLine1;
      Property CommentBlock2: String Read FCommentBlock2 Write SetCommentBlock2;
      Property CommentLine2: String Read FCommentLine2 Write FCommentLine2;
      Property HexSequence: String Read FHexSequence Write FHexSequence;
      Property IdentifierCharset: TmscrIdentifierCharset Read FIdentifierCharset Write FIdentifierCharset;
      Property SkipEOL: boolean Read FSkipEOL Write FSkipEOL;
      Property SkiPChar: String Read FSkiPChar Write FSkiPChar;
      Property SkipSpace: boolean Read FSkipSpace Write FSkipSpace;
      Property StringQuotes: String Read FStringQuotes Write FStringQuotes;
      Property SpecStrChar: boolean Read FSpecStrChar Write FSpecStrChar;
   End;

   //===========================================

Function msStrToXML(Const s: String): String;
Function msValueToXML(Const Value: variant): String;
Function msXMLToStr(Const s: String): String;

Implementation

Function msXMLToStr(Const s: String): String;
Var
   i, j, h, n: integer;
Begin
   Result := s;

   i := 1;
   n := Length(s);
   While i < n Do
   Begin
      If Result[i] = '&' Then
         If (i + 3 <= n) And (Result[i + 1] = '#') Then
         Begin
            j := i + 3;
            While Result[j] <> ';' Do
               Inc(j);
            h := StrToInt(Copy(Result, i + 2, j - i - 2));
            Delete(Result, i, j - i);
            Result[i] := Chr(h);
            Dec(n, j - i);
         End
         Else If Copy(Result, i + 1, 5) = 'quot;' Then
         Begin
            Delete(Result, i, 6);
            Result[i] := '"';
            Dec(n, 6);
         End;
      Inc(i);
   End;
End;

Function msStrToXML(Const s: String): String;
Const
   SpecChars = ['<', '>', '"', #10, #13, '&'];
Var
   i: integer;

   //...........................
   Procedure ReplaceChars(Var s: String; i: integer);
   Begin
      Insert('#' + IntToStr(Ord(s[i])) + ';', s, i + 1);
      s[i] := '&';
   End;

   //...........................
Begin
   Result := s;
   For i := Length(s) Downto 1 Do
      If s[i] In SpecChars Then
         If s[i] <> '&' Then
            ReplaceChars(Result, i)
         Else
         Begin
            If Copy(s, i + 1, 5) = 'quot;' Then
            Begin
               Delete(Result, i, 6);
               Insert('&#34;', Result, i);
            End;
         End;
End;

Function msValueToXML(Const Value: variant): String;
Begin
   Case TVarData(Value).VType Of
      varSmallint, varInteger, varByte: Result := IntToStr(integer(Value));
      varSingle, varDouble, varCurrency: Result := FloatToStr(double(Value));
      varDate: Result := DateToStr(TDateTime(Value));
      varOleStr, varString, varVariant: Result := msStrToXML(Value);

      varBoolean:
         If Value = True Then
            Result := '1'
         Else
            Result := '0';

      Else
         Result := '';
   End;
End;

//==============  TmscrDATADocument =========================

Constructor TmscrDATADocument.Create;
Begin
   FRoot := TmscrDATAItem.Create;
End;

Destructor TmscrDATADocument.Destroy;
Begin
   FRoot.Free;
   Inherited;
End;

Procedure TmscrDATADocument.Clear;
Begin
   FRoot.Clear;
End;

Procedure TmscrDATADocument.LoadFromStream(Stream: TStream);
Var
   dr: TmscrDATAReader;
Begin
   dr := TmscrDATAReader.Create(Stream);
   Try
      FRoot.Clear;
      dr.ReadHeader;
      dr.ReadRootItem(FRoot);
   Finally
      dr.Free;
   End;
End;

Procedure TmscrDATADocument.SaveToStream(Stream: TStream);
Var
   dw: TmscrDATAWriter;
Begin
   dw := TmscrDATAWriter.Create(Stream);
   dw.FAutoIndent := FAutoIndent;

   Try
      dw.WriteHeader;
      dw.WriteRootItem(FRoot);
   Finally
      dw.Free;
   End;
End;

Procedure TmscrDATADocument.LoadFromFile(Const FileName: String);
Var
   s: TFileStream;
Begin
   s := TFileStream.Create(FileName, fmOpenRead Or fmShareDenyWrite);
   Try
      LoadFromStream(s);
   Finally
      s.Free;
   End;
End;

Procedure TmscrDATADocument.SaveToFile(Const FileName: String);
Var
   s: TFileStream;
Begin
   s := TFileStream.Create(FileName, fmCreate);
   Try
      SaveToStream(s);
   Finally
      s.Free;
   End;
End;

//==============  TmscrDATAItem ============================

Procedure TmscrDATAItem.Clear;
Begin
   If FItems <> Nil Then
   Begin
      While FItems.Count > 0 Do
         TmscrDATAItem(FItems[0]).Free;
      FItems.Free;
      FItems := Nil;
   End;
End;

Destructor TmscrDATAItem.Destroy;
Begin
   Clear;
   If FParent <> Nil Then
      FParent.FItems.Remove(Self);
   Inherited;
End;

Function TmscrDATAItem.GetItems(Index: integer): TmscrDATAItem;
Begin
   Result := TmscrDATAItem(FItems[Index]);
End;

Function TmscrDATAItem.GetCount: integer;
Begin
   If FItems = Nil Then
      Result := 0
   Else
      Result := FItems.Count;
End;

Procedure TmscrDATAItem.SetParent(Const Value: TmscrDATAItem);
Begin
   If FParent <> Nil Then
      FParent.FItems.Remove(Self);
   FParent := Value;
End;

Function TmscrDATAItem.Add: TmscrDATAItem;
Begin
   Result := TmscrDATAItem.Create;
   AddItem(Result);
End;

Procedure TmscrDATAItem.InsertItem(Index: integer; Item: TmscrDATAItem);
Begin
   AddItem(Item);
   FItems.Delete(FItems.Count - 1);
   FItems.Insert(Index, Item);
End;

Procedure TmscrDATAItem.AddItem(Item: TmscrDATAItem);
Begin
   If FItems = Nil Then
      FItems := TList.Create;

   FItems.Add(Item);
   If Item.FParent <> Nil Then
      Item.FParent.FItems.Remove(Item);
   Item.FParent := Self;
End;

Function TmscrDATAItem.FindItem(Const Name: String): TmscrDATAItem;
Var
   i: integer;
Begin
   i := Find(Name);

   If i = -1 Then
   Begin
      Result := Add;
      Result.Name := Name;
   End
   Else
   Begin
      Result := Items[i];
   End;
End;

Function TmscrDATAItem.Find(Const Name: String): integer;
Var
   i: integer;
Begin
   Result := -1;
   For i := 0 To Count - 1 Do
      If AnsiCompareText(Items[i].Name, Name) = 0 Then
      Begin
         Result := i;
         break;
      End;
End;

Function TmscrDATAItem.Root: TmscrDATAItem;
Begin
   Result := Self;
   While Result.Parent <> Nil Do
      Result := Result.Parent;
End;

Function TmscrDATAItem.GetProp(Index: String): String;
Var
   i: integer;
Begin
   i := Pos(' ' + AnsiUpperCase(Index) + '="', AnsiUpperCase(' ' + FText));

   If i <> 0 Then
   Begin
      Result := Copy(FText, i + Length(Index + '="'), MaxInt);
      Result := msXMLToStr(Copy(Result, 1, Pos('"', Result) - 1));
   End
   Else
   Begin
      Result := '';
   End;
End;

Function TmscrDATAItem.PropExists(Const Index: String): boolean;
Begin
   Result := Pos(' ' + AnsiUpperCase(Index) + '="', ' ' + AnsiUpperCase(FText)) > 0;
End;

Function TmscrDATAItem.IndexOf(Item: TmscrDATAItem): integer;
Begin
   Result := FItems.IndexOf(Item);
End;

Procedure TmscrDATAItem.SetProp(Index: String; Const Value: String);
Var
   i, j: integer;
   s: String;
Begin
   i := Pos(' ' + AnsiUpperCase(Index) + '="', AnsiUpperCase(' ' + FText));

   If i <> 0 Then
   Begin
      j := i + Length(Index + '="');
      While (j <= Length(FText)) And (FText[j] <> '"') Do
         Inc(j);
      Delete(FText, i, j - i + 1);
   End
   Else
   Begin
      i := Length(FText) + 1;
   End;

   s := Index + '="' + msStrToXML(Value) + '"';
   If (i > 1) And (FText[i - 1] <> ' ') Then
      s := ' ' + s;
   Insert(s, FText, i);
End;

Procedure TmscrDATAItem.Assign(Item: TmscrDATAItem);
//..............................
   Procedure Compile_Assign(ItemFrom, ItemTo: TmscrDATAItem);
   Var
      i: integer;
   Begin
      ItemTo.Name := ItemFrom.Name;
      ItemTo.Text := ItemFrom.Text;
      ItemTo.Data := ItemFrom.Data;
      For i := 0 To ItemFrom.Count - 1 Do
         Compile_Assign(ItemFrom[i], ItemTo.Add);
   End;

   //.............................
Begin
   Clear;

   If Item <> Nil Then
      Compile_Assign(Item, Self);
End;

//==============  TmscrDATAWriter ==============

Constructor TmscrDATAWriter.Create(Stream: TStream);
Begin
   FStream := Stream;
End;

Procedure TmscrDATAWriter.FlushBuffer;
Begin
   If FBuffer <> '' Then
      FStream.Write(FBuffer[1], Length(FBuffer));
   FBuffer := '';
End;

Procedure TmscrDATAWriter.WriteLn(Const s: ansistring);
Begin
   If Not FAutoIndent Then
      Insert(s, FBuffer, MaxInt)
   Else
      Insert(s + #13#10, FBuffer, MaxInt);

   If Length(FBuffer) > 4096 Then FlushBuffer;
End;

Procedure TmscrDATAWriter.WriteHeader;
Begin
   WriteLn('<?xml version="1.0"?>');
End;

Function Dup(n: integer): ansistring;
Begin
   SetLength(Result, n);
   FillChar(Result[1], n, ' ');
End;

Procedure TmscrDATAWriter.WriteItem(Item: TmscrDATAItem; Level: integer = 0);
Var
   s: ansistring;
Begin
   If Item.FText <> '' Then
   Begin
      s := Item.FText;
      If (s = '') Or (s[1] <> ' ') Then
         s := ' ' + s;

   End
   Else
   Begin
      s := '';
   End;

   If Item.Count = 0 Then
      s := s + '/>'
   Else
      s := s + '>';

   If Not FAutoIndent Then
      s := '<' + ansistring(Item.Name) + s
   Else
      s := Dup(Level) + '<' + ansistring(Item.Name) + s;

   WriteLn(s);
End;

Procedure TmscrDATAWriter.WriteRootItem(RootItem: TmscrDATAItem);
//............................
   Procedure DoWrite(RootItem: TmscrDATAItem; Level: integer = 0);
   Var
      i: integer;
      NeedClear: boolean;
   Begin
      NeedClear := False;
      If Not FAutoIndent Then
         Level := 0;

      WriteItem(RootItem, Level);
      For i := 0 To RootItem.Count - 1 Do
         DoWrite(RootItem[i], Level + 2);

      If RootItem.Count > 0 Then
         If Not FAutoIndent Then
            WriteLn('</' + ansistring(RootItem.Name) + '>')
         Else
            WriteLn(Dup(Level) + '</' + ansistring(RootItem.Name) + '>');

      If NeedClear Then
         RootItem.Clear;
   End;

   //..............................
Begin
   DoWrite(RootItem);
   FlushBuffer;
End;

//==============  TmscrDATAReader =========================

Constructor TmscrDATAReader.Create(Stream: TStream);
Begin
   FStream := Stream;
   FSize := Stream.Size;
   FPosition := Stream.Position;
   GetMem(FBuffer, 4096);
End;

Destructor TmscrDATAReader.Destroy;
Begin
   FreeMem(FBuffer, 4096);
   FStream.Position := FPosition;
   Inherited;
End;

Procedure TmscrDATAReader.ReadBuffer;
Begin
   FBufEnd := FStream.Read(FBuffer^, 4096);
   FBufPos := 0;
End;

Procedure TmscrDATAReader.SetPosition(Const Value: int64);
Begin
   FPosition := Value;
   FStream.Position := Value;
   FBufPos := 0;
   FBufEnd := 0;
End;

Procedure TmscrDATAReader.RaiseException;
Begin
   Raise Exception.Create('Invalid file format');
End;

Procedure TmscrDATAReader.ReadHeader;
Var
   s1: String;
   s2: String;
Begin
   ReadItem(s1, s2);
   If Pos('?xml', s1) <> 1 Then
      RaiseException;
End;

Procedure TmscrDATAReader.ReadItem(Var Name: String; Var Text: String);
Var
   ps: pansichar;
   cb: byte;
   i, ic: integer;
   curpos, len: integer;
   xstate: (FindLeft, FindRight, FindComment, Done);
Begin
   Text := '';
   ic := 0;
   xstate := FindLeft;
   curpos := 0;
   len := 4096;
   SetLength(Name, len);
   ps := @Name[1];

   While FPosition < FSize Do
   Begin
      If FBufPos = FBufEnd Then
         ReadBuffer;
      cb := Ord(FBuffer[FBufPos]);
      Inc(FBufPos);
      Inc(FPosition);

      If xstate = FindLeft Then
      Begin
         If cb = Ord('<') Then
            xstate := FindRight;
      End
      Else If xstate = FindRight Then
      Begin

         If cb = Ord('>') Then
         Begin
            xstate := Done;
            break;
         End
         Else If cb = Ord('<') Then
            RaiseException
         Else
         Begin
            ps[curpos] := ansichar(Chr(cb));
            Inc(curpos);
            If (curpos = 3) And (Pos(ansistring('!--'), Name) = 1) Then
            Begin
               xstate := FindComment;
               ic := 0;
               curpos := 0;
            End;
            If curpos >= len - 1 Then
            Begin
               Inc(len, 4096);
               SetLength(Name, len);
               ps := @Name[1];
            End;
         End;

      End
      Else If xState = FindComment Then
      Begin
         If ic = 2 Then
         Begin

            If cb = Ord('>') Then
               xstate := FindLeft
            Else
               ic := 0;

         End
         Else
         Begin

            If cb = Ord('-') Then
               Inc(ic)
            Else
               ic := 0;

         End;
      End;
   End;

   len := curpos;
   SetLength(Name, len);

   If xstate = FindRight Then RaiseException;

   If (Name <> '') And (Name[len] = ' ') Then
      SetLength(Name, len - 1);

   i := Pos(ansistring(' '), Name);
   If i <> 0 Then
   Begin
      Text := Copy(Name, i + 1, len - i);
      Delete(Name, i, len - i + 1);
   End;
End;

Procedure TmscrDATAReader.ReadRootItem(Item: TmscrDATAItem);
Var
   xLastName: String;

   //........................................
   Function DoRead(RootItem: TmscrDATAItem): boolean;
   Var
      n: integer;
      ChildItem: TmscrDATAItem;
      Done: boolean;
   Begin
      Result := False;
      ReadItem(RootItem.FName, RootItem.FText);
      xLastName := RootItem.FName;

      If (RootItem.Name = '') Or (RootItem.Name[1] = '/') Then
      Begin
         Result := True;
         Exit;
      End;

      n := Length(RootItem.Name);
      If RootItem.Name[n] = '/' Then
      Begin
         SetLength(RootItem.FName, n - 1);
         Exit;
      End;

      n := Length(RootItem.Text);
      If (n > 0) And (RootItem.Text[n] = '/') Then
      Begin
         SetLength(RootItem.FText, n - 1);
         Exit;
      End;

      Repeat
         ChildItem := TmscrDATAItem.Create;
         Done := DoRead(ChildItem);
         If Not Done Then
            RootItem.AddItem(ChildItem)
         Else
            ChildItem.Free;
      Until Done;

      If (xLastName <> '') And (AnsiCompareText(xLastName, '/' + RootItem.Name) <> 0) Then
         RaiseException;
   End;

   //.............................
Begin
   DoRead(Item);
End;

//=============== TmscrParser ==========================================

Constructor TmscrParser.Create;
Begin
   FKeywords := TStringList.Create;
   TStringList(FKeywords).Sorted := True;
   FYList := TList.Create;
   FUseY := True;
   Clear;
End;

Destructor TmscrParser.Destroy;
Begin
   FKeywords.Free;
   FYList.Free;
   Inherited;
End;

Procedure TmscrParser.Clear;
Begin
   FSpecStrChar := False;
   FCommentLine1 := '//';
   CommentBlock1 := '{,}';
   CommentBlock2 := '(*,*)';
   FHexSequence := '$';
   FIdentifierCharset := ['_', '0'..'9', 'a'..'z', 'A'..'Z'];
   FSkipChar := '';
   FSkipEOL := True;
   FStringQuotes := '''';
   FSkipSpace := True;
   FCaseSensitive := False;
End;

Procedure TmscrParser.InitKeywords;
Begin
   FKeywords.Clear;
   With FKeywords Do
   Begin
      Add('and');
      Add('array');
      Add('begin');
      Add('break');
      Add('case');
      Add('const');
      Add('continue');
      Add('div');
      Add('do');
      Add('downto');
      Add('else');
      Add('end');
      Add('except');
      Add('exit');
      Add('finally');
      Add('for');
      Add('function');
      Add('goto');
      Add('if');
      Add('in');
      Add('is');
      Add('label');
      Add('mod');
      Add('not');
      Add('of');
      Add('or');
      Add('procedure');
      Add('program');
      Add('repeat');
      Add('shl');
      Add('shr');
      Add('then');
      Add('to');
      Add('try');
      Add('until');
      Add('uses');
      Add('var');
      Add('while');
      Add('with');
      Add('xor');
   End;

End;

Procedure TmscrParser.SetCommentBlock1(Const Value: String);
Var
   sl: TStringList;
Begin
   FCommentBlock1 := Value;
   FCommentBlock11 := '';
   FCommentBlock12 := '';

   sl := TStringList.Create;
   sl.CommaText := FCommentBlock1;
   If sl.Count > 0 Then
      FCommentBlock11 := sl[0];
   If sl.Count > 1 Then
      FCommentBlock12 := sl[1];
   sl.Free;
End;

Procedure TmscrParser.SetCommentBlock2(Const Value: String);
Var
   sl: TStringList;
Begin
   FCommentBlock2 := Value;
   FCommentBlock21 := '';
   FCommentBlock22 := '';

   sl := TStringList.Create;
   sl.CommaText := FCommentBlock2;
   If sl.Count > 0 Then
      FCommentBlock21 := sl[0];
   If sl.Count > 1 Then
      FCommentBlock22 := sl[1];

   sl.Free;
End;

Procedure TmscrParser.SetPosition(Const Value: integer);
Begin
   FPosition := Value;
   FLastPosition := Value;
End;

Procedure TmscrParser.SetText(Const Value: String);
Var
   i: integer;
Begin
   FText := Value + #0;
   FLastPosition := 1;
   FPosition := 1;
   FSize := Length(Value);

   If FUseY Then
   Begin
      FYList.Clear;
      FYList.Add(TObject(0));
      For i := 1 To FSize Do
         If FText[i] = #10 Then
            FYList.Add(TObject(PtrInt(i)));
   End;
End;

Procedure TmscrParser.ConstructCharset(Const s: String);
Var
   i: integer;
Begin
   FIdentifierCharset := [];
   For i := 1 To Length(s) Do
      FIdentifierCharset := FIdentifierCharset + [s[i]];
End;

Function TmscrParser.GetEOL: boolean;
Begin
   SkipSpaces;

   If FText[FPosition] In [#10, #13] Then
   Begin
      Result := True;
      While FText[FPosition] In [#10, #13] Do
         Inc(FPosition);
   End
   Else
   Begin
      Result := False;
   End;
End;

Procedure TmscrParser.SkipSpaces;
Var
   s1, s2: String;
   Flag, CLine: boolean;
   Spaces: Set Of ansichar;
Begin
   Spaces := [#0..#32];
   If Not FSkipEOL Then

{$IFDEF Windows}
      Spaces := Spaces - [#13];
{$ELSE}
      Spaces := Spaces - [#10];
{$ENDIF}

   While (FPosition <= FSize) And (FText[FPosition] In Spaces) Do
      Inc(FPosition);

   If (FPosition <= FSize) And (FSkipChar <> '') And (FText[FPosition] = FSkipChar[1]) Then
   Begin
      Inc(FPosition);
      GetEOL;
      SkipSpaces;
   End;

   If FPosition < FSize Then
   Begin
      If FCommentLine1 <> '' Then
         s1 := Copy(FText, FPosition, Length(FCommentLine1))
      Else
         s1 := ' ';

      If FCommentLine2 <> '' Then
         s2 := Copy(FText, FPosition, Length(FCommentLine2))
      Else
         s2 := ' ';

      If (s1 = FCommentLine1) Or (s2 = FCommentLine2) Then
      Begin
         CLine := (FPosition - 1 > 0) And (FText[FPosition - 1] <> #10) And Not FSkipEOL;
         While (FPosition <= FSize) And (FText[FPosition] <> #10) Do
         Begin
            If (FText[FPosition] =
{$IFDEF UNIX}
               #10
{$ELSE}
               #13
{$ENDIF}
               ) And CLine Then break;
            Inc(FPosition);
         End;
         SkipSpaces;
      End
      Else
      Begin
         Flag := False;

         If FCommentBlock1 <> '' Then
         Begin
            s1 := Copy(FText, FPosition, Length(FCommentBlock11));
            If s1 = FCommentBlock11 Then
            Begin
               Flag := True;
               s2 := FCommentBlock12;
            End;
         End;

         If Not Flag And (FCommentBlock2 <> '') Then
         Begin
            s1 := Copy(FText, FPosition, Length(FCommentBlock21));
            If s1 = FCommentBlock21 Then
            Begin
               Flag := True;
               s2 := FCommentBlock22;
            End;
         End;

         If Flag Then
         Begin
            Inc(FPosition, Length(s2));
            While (FPosition <= FSize) And (Copy(FText, FPosition, Length(s2)) <> s2) Do
               Inc(FPosition);
            Inc(FPosition, Length(s2));
            SkipSpaces;
         End;

      End;
   End;

   FLastPosition := FPosition;
End;

Function TmscrParser.Ident: String;
Begin
   If FSkipSpace Then
      SkipSpaces;

   If (FText[FPosition] In FIdentifierCharset) And Not (FText[FPosition] In ['0'..'9']) Then
   Begin
      While (FText[FPosition] In FIdentifierCharset) Do
         Inc(FPosition);

      Result := Copy(FText, FLastPosition, FPosition - FLastPosition);
   End
   Else
   Begin
      Result := '';
   End;
End;

Function TmscrParser.IsKeyWord(Const s: String): boolean;
Var
   i: integer;
Begin
   If FCaseSensitive Then
   Begin
      Result := False;
      For i := 0 To FKeywords.Count - 1 Do
      Begin
         Result := FKeywords[i] = s;
         If Result Then break;
      End;
   End
   Else
   Begin
      Result := FKeywords.IndexOf(s) <> -1;
   End;
End;

Function TmscrParser.GetIdent: String;
Begin
   Result := Ident;
   If IsKeyWord(Result) Then
      Result := '';
End;

Function TmscrParser.GetWord: String;
Begin
   Result := Ident;
End;

Function TmscrParser.GetChar: String;
Begin
   If FText[FPosition] In ['!', '@', '#', '$', '%', '^', '&', '|', '\', '.', ',', ':', ';', '?', '''', '"', '~', '`', '_', '[', ']', '{', '}', '(', ')', '+', '-', '*', '/', '=', '<', '>'] Then

   Begin
      Result := FText[FPosition];
      Inc(FPosition);
   End
   Else
   Begin
      Result := '';
   End;
End;

Function TmscrParser.GetString: String;
Var
   Flag: boolean;
   Str: String;
   FError: boolean;
   FCpp: boolean;

   //....................................
   Function DoQuotedString: boolean;
   Var
      i, j: integer;
   Begin
      Result := False;
      i := FPosition;

      If FText[FPosition] = FStringQuotes[1] Then
      Begin
         Repeat
            Inc(FPosition);

            If FCpp And (FText[FPosition] = '\') Then
            Begin

               Case Lowercase(FText[FPosition + 1]) Of
                  'n':
                     Begin
                        Str := Str + #10;
                        Inc(FPosition);
                     End;

                  'r':
                     Begin
                        Str := Str + #13;
                        Inc(FPosition);
                     End;

                  'x':
                     Begin
                        Inc(FPosition, 2);
                        j := FPosition;
                        Result := DoHexDigitSequence;

                        If Result Then
                           Str := Str + Chr(StrToInt('$' + Copy(FText, j, FPosition - j)))
                        Else
                           FPosition := j;

                        Dec(FPosition);
                     End;

                  Else
                     Begin
                        Str := Str + FText[FPosition + 1];
                        Inc(FPosition);
                     End;
               End;

            End
            Else If FText[FPosition] = FStringQuotes[1] Then
            Begin

               If Not FCpp And (FText[FPosition + 1] = FStringQuotes[1]) Then
               Begin
                  Str := Str + FStringQuotes[1];
                  Inc(FPosition);
               End
               Else
               Begin
                  break;
               End;

            End
            Else
            Begin
               Str := Str + FText[FPosition];
            End;

         Until FText[FPosition] In [#0..#31] - [#9];

         If FText[FPosition] = FStringQuotes[1] Then
         Begin
            Inc(FPosition);
            Result := True;
         End
         Else
         Begin
            FPosition := i;
         End;

      End;
   End;

   //..................................
   Function DoControlString: boolean;
   Var
      i: integer;
   Begin
      Result := False;
      i := FPosition;

      If FText[FPosition] = '#' Then
      Begin
         Inc(FPosition);
         Result := DoUnsignedInteger;
         If Result Then
            Str := Chr(StrToInt(Copy(FText, i + 1, FPosition - i - 1)))
         Else
            FPosition := i;
      End;
   End;

   //..................................

{$HINTS OFF}
Begin
   Result := '';
   If FSkipSpace Then
      SkipSpaces;
   Flag := True;
   FError := False;
   FCpp := {FStringQuotes = '"'} FSpecStrChar;

   Repeat
      Str := '';
      If DoQuotedString Or DoControlString Then
         Result := Result + Str
      Else
      Begin
         FError := Flag;
         break;
      End;

      Flag := False;
   Until False;

   If Not FError Then
      Result := '''' + Result + '''';
End;

{$HINTS ON}

Function TmscrParser.DoDigitSequence: boolean;
Begin
   Result := False;

   If FText[FPosition] In ['0'..'9'] Then
   Begin
      While FText[FPosition] In ['0'..'9'] Do
         Inc(FPosition);
      Result := True;
   End;
End;

Function TmscrParser.DoHexDigitSequence: boolean;
Begin
   Result := False;

   If FText[FPosition] In ['0'..'9', 'a'..'f', 'A'..'F'] Then
   Begin
      While FText[FPosition] In ['0'..'9', 'a'..'f', 'A'..'F'] Do
         Inc(FPosition);
      Result := True;
   End;
End;

Function TmscrParser.DoUnsignedInteger: boolean;
Var
   Pos1: integer;
   s: String;
Begin
   Pos1 := FPosition;
   s := Copy(FText, FPosition, Length(FHexSequence));

   If s = FHexSequence Then
   Begin
      Inc(FPosition, Length(s));
      Result := DoHexDigitSequence;
   End
   Else
   Begin
      Result := DoDigitSequence;
   End;

   If Not Result Then
      FPosition := Pos1;
End;

Function TmscrParser.DoUnsignedReal: boolean;
Var
   Pos1, Pos2: integer;
Begin
   Pos1 := FPosition;
   Result := DoUnsignedInteger;

   If Result Then
   Begin
      If FText[FPosition] = '.' Then
      Begin
         Inc(FPosition);
         Result := DoDigitSequence;
      End;

      If Result Then
      Begin
         Pos2 := FPosition;
         If Not DoScaleFactor Then
            FPosition := Pos2;
      End;
   End;

   If Not Result Then
      FPosition := Pos1;
End;

Function TmscrParser.DoScaleFactor: boolean;
Begin
   Result := False;

   If FText[FPosition] In ['e', 'E'] Then
   Begin
      Inc(FPosition);

      If FText[FPosition] In ['+', '-'] Then
         Inc(FPosition);

      Result := DoDigitSequence;
   End;
End;

Function TmscrParser.GetNumber: String;
Var
   Pos1: integer;
Begin
   Result := '';
   If FSkipSpace Then
      SkipSpaces;
   Pos1 := FPosition;

   If DoUnsignedReal Or DoUnsignedInteger Then
      Result := Copy(FText, FLastPosition, FPosition - FLastPosition)
   Else
      FPosition := Pos1;

   If FHexSequence <> '$' Then
      While Pos(FHexSequence, Result) <> 0 Do
      Begin
         Pos1 := Pos(FHexSequence, Result);
         Delete(Result, Pos1, Length(FHexSequence));
         Insert('$', Result, Pos1);
      End;
End;

Function TmscrParser.GetFRString: String;
Var
   i, c: integer;
   fl1, fl2: boolean;
Begin
   Result := '';
   i := FPosition;
   fl1 := True;
   fl2 := True;
   c := 1;

   Dec(FPosition);

   Repeat
      Inc(FPosition);

      If fl1 And fl2 Then

         If FText[FPosition] In ['<', '['] Then
            Inc(c)
         Else If FText[FPosition] In ['>', ']'] Then
            Dec(c);

      If fl1 Then
         If FText[FPosition] = '"' Then
            fl2 := Not fl2;

      If fl2 Then
         If FText[FPosition] = '''' Then
            fl1 := Not fl1;

   Until (c = 0) Or (FPosition >= Length(FText));

   Result := Copy(FText, i, FPosition - i);
End;

Function TmscrParser.GetXYPosition: String;
Var
   i, i0, i1, c, pos, X, Y: integer;
Begin
   i0 := 0;
   i1 := FYList.Count - 1;

   While i0 <= i1 Do
   Begin
      i := (i0 + i1) Div 2;
      pos := integer(FYList[i]);

      If pos = FPosition Then
         c := 0
      Else If pos > FPosition Then
         c := 1
      Else
         c := -1;

      If c < 0 Then
         i0 := i + 1
      Else
      Begin
         i1 := i - 1;
         If c = 0 Then
            i0 := i;
      End;
   End;

   X := 1;
   Y := i0;
   i := integer(FYList[i0 - 1]) + 1;

   While i < FPosition Do
   Begin
      Inc(i);
      Inc(X);
   End;
   Result := IntToStr(Y) + ':' + IntToStr(X);
End;

Function TmscrParser.GetPlainPosition(pt: TPoint): integer;
Var
   i: integer;
Begin
   Result := -1;
   i := pt.Y - 1;
   If (i >= 0) And (i < FYList.Count) Then
      Result := integer(FYList[i]) + pt.X;
End;

//=========================================================

End.
