{*************************************************************************
                Copyright (c) PilotLogic Software House

  Package pl_MagicScript
  This file is part of CodeTyphon Studio (https://www.pilotlogic.com/)

 ***** BEGIN LICENSE BLOCK *****
 * Version: LGPL-2.1 with link exception (Modified LGPL)
 *
 * The contents of this file are subject to the
 * GNU LIBRARY GENERAL PUBLIC LICENSE Version 2.1 (the "License")
 * with the following modification:
 *
 * As a special exception, the copyright holders of this library give you
 * permission to link this library with independent modules to produce an
 * executable, regardless of the license terms of these independent modules,
 * and to copy and distribute the resulting executable under terms of your choice,
 * provided that you also meet, for each linked independent module, the terms
 * and conditions of the license of that module. An independent module is a
 * module which is not derived from or based on this library. If you modify this
 * library, you may extend this exception to your version of the library, but
 * you are not obligated to do so. If you do not wish to do so, delete this
 * exception statement from your version.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * ***** END LICENSE BLOCK *****
 
***************************************************************************}

Unit msLibrariesTree;

Interface

{$I magicscript.inc}

Uses
   SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
   Rtti, TypInfo,
   StdCtrls, ExtCtrls, ComCtrls, SynEdit,
   msbaseobjects, mscoreengine;

Type

   TmscrRTTILibrariesTree = Class(TPanel)
   Private
      FDATAList: TmscrDATADocument;
      FScript: TmscrScript;
      FTreeView: TTreeView;
      FTreeView2: TTreeView;
      Splitter1: TSplitter;
      FImages: TImageList;
      FShowFunctions: boolean;
      FShowClasses: boolean;
      FShowRecords: boolean; 
      FShowConstants: boolean;
      FShowTypes: boolean;
      FShowVariables: boolean;
      FExpanded: boolean;
      FExpandLevel: integer;
      FScriptEditor: TSynEdit;
      FUpdating: boolean;
      FDualPanel: boolean;
      Procedure SetScriptEditor(Value: TSynEdit);
      Procedure SetImageIndex(Node: TTreeNode; Index: integer);
      Procedure SetScript(Const Value: TmscrScript);
      Procedure SetDualPanel(Const Value: boolean);
   Protected
      Procedure Notification(AComponent: TComponent; Operation: TOperation); Override;
      Function GetCategoryByName(s: String): String;

      Procedure FindAllClasses(xi: TmscrDATAItem; Root: TTreeNode);
      Procedure FindAllRecords(xi: TmscrDATAItem; Root: TTreeNode);
      Procedure FindAllFuncProc(xi: TmscrDATAItem; Root: TTreeNode);
      Procedure FindAllConstants(xi: TmscrDATAItem; Root: TTreeNode);
      Procedure FindAllTypes(xi: TmscrDATAItem; Root: TTreeNode);
      Procedure FindAllVariables(xi: TmscrDATAItem; Root: TTreeNode);

      Procedure FindAllClasses2(xi: TmscrDATAItem; Root: TTreeNode);
      Procedure FindAllRecords2(xi: TmscrDATAItem; Root: TTreeNode);
      Procedure FindAllFuncProc2(xi: TmscrDATAItem; Root: TTreeNode);
      Procedure FindAllConstants2(xi: TmscrDATAItem; Root: TTreeNode);
      Procedure FindAllTypes2(xi: TmscrDATAItem; Root: TTreeNode);
      Procedure FindAllVariables2(xi: TmscrDATAItem; Root: TTreeNode);

      Procedure TreeViewBuild;
      Procedure TreeViewExpandNodes(level: integer);
      Procedure TreeChange(Sender: TObject; Node: TTreeNode);

      Procedure DoOnTreeViewClick(Sender: TObject);
      Procedure DoOnTreeViewDblClick(Sender: TObject);
      Procedure DoOnTreeViewCollapsed(Sender: TObject; Node: TTreeNode);
      Procedure DoOnTreeViewExpanded(Sender: TObject; Node: TTreeNode);
   Public
      Constructor Create(AOwner: TComponent); Override;
      Destructor Destroy; Override;
      Procedure UpdateItems;
      Function GetFieldName: String;

   Published
      Property Script: TmscrScript Read FScript Write SetScript;
      Property ScriptEditor: TSynEdit Read FScriptEditor Write SetScriptEditor;

      Property ShowClasses: boolean Read FShowClasses Write FShowClasses;
      Property ShowRecords: boolean Read FShowRecords Write FShowRecords;
      Property ShowFunctions: boolean Read FShowFunctions Write FShowFunctions; 
      Property ShowConstants: boolean Read FShowConstants Write FShowConstants;
      Property ShowTypes: boolean Read FShowTypes Write FShowTypes;
      Property ShowVariables: boolean Read FShowVariables Write FShowVariables;
      Property Expanded: boolean Read FExpanded Write FExpanded;
      Property ExpandLevel: integer Read FExpandLevel Write FExpandLevel;
      Property DualPanel: boolean Read FDualPanel Write SetDualPanel;

      Property Align;
      Property Anchors;
      Property DragCursor;
      Property DragKind;
      Property DragMode;
      Property BiDiMode;
      Property ParentBiDiMode;
      Property Color;
      Property Constraints;
      Property Enabled;
      Property Font;
      Property ParentColor;
      Property ParentFont;
      Property ParentShowHint;
      Property PopupMenu;
      Property ShowHint;
      Property TabOrder;
      Property TabStop;
      Property Visible;
      Property OnClick;
      Property OnDblClick;
      Property OnDragDrop;
      Property OnDragOver;
      Property OnEndDrag;
      Property OnEnter;
      Property OnExit;
      Property OnKeyDown;
      Property OnMouseDown;
      Property OnMouseMove;
      Property OnMouseUp;
      Property OnStartDrag;
   End;

Implementation

{$R *.res}

Const
   cnImgBook = 0;
   cnImgBookOpen = 1;
   cnImgBookGray = 2;
   cnImgGlobalfunction = 3;
   cnImgProcedure = 4;
   cnImgFunction = 5;
   cnImgEvent = 6;
   cnImgElement = 7;
   cnImgVariable = 8;
   cnImgType = 9;
   cnImgProperty = 10;
   cnImgConstructor = 11;   
   cnImgConstant = 12;

Constructor TmscrRTTILibrariesTree.Create(AOwner: TComponent);
Begin
   Inherited;
   Parent := AOwner As TWinControl;
   BevelOuter := bvNone;
   //...........
   FTreeView := TTreeView.Create(Self);
   FTreeView.Parent := Self;
   FTreeView.Align := alClient;
   FTreeView.DragMode := dmAutomatic;
   FTreeView.ReadOnly := True;
   FTreeView.TabOrder := 0;

   FTreeView2 := TTreeView.Create(Self);
   FTreeView2.Parent := Self;
   FTreeView2.Align := alRight;
   FTreeView2.SetBounds(0, 0, 200, 100);
   FTreeView2.DragMode := dmAutomatic;
   FTreeView2.ReadOnly := True;
   FTreeView2.TabOrder := 1;

   Splitter1 := TSplitter.Create(Self);
   Splitter1.Parent := Self;
   Splitter1.Align := alRight;

   //...........
   FImages := TImageList.Create(Self);
   FImages.Masked := True;
   FImages.AddResourceName(hInstance, 'mcrpic_book');
   FImages.AddResourceName(hInstance, 'mcrpic_bookopen');
   FImages.AddResourceName(hInstance, 'mcrpic_bookgray');
   FImages.AddResourceName(hInstance, 'mcrpic_globalfunction');
   FImages.AddResourceName(hInstance, 'mcrpic_procedure');
   FImages.AddResourceName(hInstance, 'mcrpic_function');
   FImages.AddResourceName(hInstance, 'mcrpic_event');
   FImages.AddResourceName(hInstance, 'mcrpic_element');
   FImages.AddResourceName(hInstance, 'mcrpic_variable');
   FImages.AddResourceName(hInstance, 'mcrpic_type');
   FImages.AddResourceName(hInstance, 'mcrpic_property');
   FImages.AddResourceName(hInstance, 'mcrpic_constructor');
   FImages.AddResourceName(hInstance, 'mcrpic_constant');

   FTreeView.Images := FImages;
   FTreeView.OnCollapsed := DoOnTreeViewCollapsed;
   FTreeView.OnExpanded := DoOnTreeViewExpanded;
   FTreeView.OnClick := DoOnTreeViewClick;
   FTreeView.OnDblClick := DoOnTreeViewDblClick;

   FTreeView2.Images := FImages;
   FTreeView2.OnCollapsed := DoOnTreeViewCollapsed;
   FTreeView2.OnExpanded := DoOnTreeViewExpanded;
   FTreeView2.OnDblClick := DoOnTreeViewDblClick;

   FDualPanel := True;
   FDATAList := TmscrDATADocument.Create;
   Expanded := True;
   ExpandLevel := 2;

   ShowClasses := True;
   ShowRecords := True;
   ShowFunctions := True;  
   ShowConstants := True;
   ShowTypes := True;
   ShowVariables := True;

   UpdateItems;
   Height := 250;
   Width := 200;
End;

Destructor TmscrRTTILibrariesTree.Destroy;
Begin
   FImages.Free;
   FTreeView.Free;
   FUpdating := True;
   FDATAList.Free;
   Inherited;
End;

Procedure TmscrRTTILibrariesTree.SetImageIndex(Node: TTreeNode; Index: integer);
Begin
   Node.ImageIndex := Index;
   Node.StateIndex := Index;
   Node.SelectedIndex := Index;
End;

Procedure TmscrRTTILibrariesTree.SetScriptEditor(Value: TSynEdit);
Begin
   FScriptEditor := Value;
End;

Procedure TmscrRTTILibrariesTree.SetScript(Const Value: TmscrScript);
Begin
   FScript := Value;
   UpdateItems;
End;

Procedure TmscrRTTILibrariesTree.SetDualPanel(Const Value: boolean);
Begin
   FDualPanel := Value;

   If FDualPanel Then
   Begin
      Splitter1.Visible := True;
      FTreeView2.Visible := True;
      FTreeView.OnClick := DoOnTreeViewClick;
   End
   Else
   Begin
      Splitter1.Visible := False;
      FTreeView2.Visible := False;
      FTreeView.OnClick := Nil;
   End;

   UpdateItems;
   Invalidate;
End;

Procedure TmscrRTTILibrariesTree.UpdateItems;
Begin
   TreeViewBuild;
End;

Procedure TmscrRTTILibrariesTree.TreeChange(Sender: TObject; Node: TTreeNode);
Begin
   If FUpdating Then
      Exit;
End;

Procedure TmscrRTTILibrariesTree.DoOnTreeViewDblClick(Sender: TObject);
Begin
   If Assigned(ScriptEditor) Then
   Begin
      If Sender = FTreeView Then
      Begin
         If FTreeView.Selected.Count = 0 Then
            ScriptEditor.SelText := FTreeView.Selected.Text;
      End;

      If Sender = FTreeView2 Then
      Begin
         If FTreeView2.Selected.Count = 0 Then
            ScriptEditor.SelText := FTreeView2.Selected.Text;
      End;
   End;

   If Assigned(OnDblClick) Then OnDblClick(Self);
End;

Procedure TmscrRTTILibrariesTree.DoOnTreeViewCollapsed(Sender: TObject; Node: TTreeNode);
Begin
   SetImageIndex(Node, cnImgBook);
End;

Procedure TmscrRTTILibrariesTree.DoOnTreeViewExpanded(Sender: TObject; Node: TTreeNode);
Begin
   SetImageIndex(Node, cnImgBookOpen);
End;

Procedure TmscrRTTILibrariesTree.Notification(AComponent: TComponent; Operation: TOperation);
Begin
   Inherited Notification(AComponent, Operation);
   If Operation = opRemove Then
      If AComponent = FScript Then
         FScript := Nil
      Else If AComponent = FScriptEditor Then
         FScriptEditor := Nil;
End;

Function TmscrRTTILibrariesTree.GetFieldName: String;
Var
   i, n: integer;
   s: String;
Begin
   Result := '';

   If (FTreeView.Selected <> Nil) And (FTreeView.Selected.Count = 0) Then
   Begin
      s := FTreeView.Selected.Text;

      If Pos('(', s) <> 0 Then
         n := 1
      Else
         n := 0;

      For i := 1 To Length(s) Do
         If s[i] In [',', ';'] Then
            Inc(n);

      If n = 0 Then
      Begin
         s := Copy(s, 1, Pos(':', s) - 1);
      End
      Else
      Begin
         s := Copy(s, 1, Pos('(', s));

         For i := 1 To n - 1 Do
            s := s + ',';

         s := s + ')';
      End;

      Result := s;
   End;
End;

//======================================================================
//======================================================================

Function TmscrRTTILibrariesTree.GetCategoryByName(s: String): String;
Begin
   Result := s;

   If SameText(s, 'catConv') Then
   Begin
      Result := 'Conversion';
      exit;
   End;
   If SameText(s, 'catFormat') Then
   Begin
      Result := 'Formatting';
      exit;
   End;
   If SameText(s, 'catDate') Then
   Begin
      Result := 'Date/Time';
      exit;
   End;
   If SameText(s, 'catString') Then
   Begin
      Result := 'String Routines';
      exit;
   End;
   If SameText(s, 'catMath') Then
   Begin
      Result := 'Math';
      exit;
   End;
   If SameText(s, 'catOther') Then
   Begin
      Result := 'Other';
      exit;
   End;

End;

Procedure TmscrRTTILibrariesTree.FindAllClasses(xi: TmscrDATAItem; Root: TTreeNode);
Var
   i: integer;
   xNode: TTreeNode;
   s: String;
Begin
   s := xi.Prop['text'];
   xNode := FTreeView.Items.AddChild(Root, s);
   xNode.Data := xi;

   If Root = Nil Then
      xNode.Text := xi.Name + ' (Total: ' + IntToStr(xi.Count) + ')';

   Case xNode.Level Of
      0, 1:
         Begin
            If xi.Count = 0 Then
               SetImageIndex(xNode, cnImgBookGray)
            Else
               SetImageIndex(xNode, cnImgBook);
         End;
      2:
         Begin
            If pos('Constructor', xNode.Text) > 0 Then
               SetImageIndex(xNode, cnImgConstructor)
            Else If pos('procedure', xNode.Text) > 0 Then
               SetImageIndex(xNode, cnImgProcedure)
            Else If pos('function', xNode.Text) > 0 Then
               SetImageIndex(xNode, cnImgFunction)
            Else If pos('event', xNode.Text) > 0 Then
               SetImageIndex(xNode, cnImgEvent)
            Else
               SetImageIndex(xNode, cnImgProperty);
         End;
   End;

   For i := 0 To xi.Count - 1 Do
      FindAllClasses(xi[i], xNode);
End;

Procedure TmscrRTTILibrariesTree.FindAllRecords(xi: TmscrDATAItem; Root: TTreeNode);
Var
   i: integer;
   xNode: TTreeNode;
   s: String;
Begin
   s := xi.Prop['text'];
   xNode := FTreeView.Items.AddChild(Root, s);
   xNode.Data := xi;

   If Root = Nil Then
      xNode.Text := xi.Name + ' (Total: ' + IntToStr(xi.Count) + ')';

   Case xNode.Level Of
      0, 1:
         Begin
            If xi.Count = 0 Then
               SetImageIndex(xNode, cnImgBookGray)
            Else
               SetImageIndex(xNode, cnImgBook);
         End;
      2:
         Begin
            SetImageIndex(xNode, cnImgElement);
         End;
   End;

   For i := 0 To xi.Count - 1 Do
      FindAllRecords(xi[i], xNode);
End;

Procedure TmscrRTTILibrariesTree.FindAllFuncProc(xi: TmscrDATAItem; Root: TTreeNode);
Var
   i: integer;
   xNode: TTreeNode;
   s: String;
Begin
   s := xi.Prop['text'];
   If xi.Count = 0 Then
      s := Copy(s, Pos(' ', s) + 1, 255)
   Else
      s := GetCategoryByName(s);

   xNode := FTreeView.Items.AddChild(Root, s);
   xNode.Data := xi;

   If Root = Nil Then
      xNode.Text := xi.Name + ' (Total: ' + IntToStr(xi.Count) + ')';

   Case xNode.Level Of
      0, 1:
         Begin
            If xi.Count = 0 Then
               SetImageIndex(xNode, cnImgBookGray)
            Else
               SetImageIndex(xNode, cnImgBook);
         End;
      2:
         Begin
            If (pos('):', xNode.Text) > 0) Or ((pos('(', xNode.Text) < 1) And (pos(':', xNode.Text) > 0)) Then
               SetImageIndex(xNode, cnImgFunction)
            Else
               SetImageIndex(xNode, cnImgProcedure);
         End;
   End;

   For i := 0 To xi.Count - 1 Do
      FindAllFuncProc(xi[i], xNode);
End;

Procedure TmscrRTTILibrariesTree.FindAllConstants(xi: TmscrDATAItem; Root: TTreeNode);
Var
   i: integer;
   xNode: TTreeNode;
   s: String;
Begin
   s := xi.Prop['text'];
   xNode := FTreeView.Items.AddChild(Root, s);
   xNode.Data := xi;

   If Root = Nil Then
      xNode.Text := xi.Name + ' (Total: ' + IntToStr(xi.Count) + ')';

   Case xNode.Level Of
      0:
         Begin
            If xi.Count = 0 Then
               SetImageIndex(xNode, cnImgBookGray)
            Else
               SetImageIndex(xNode, cnImgBook);
         End;
      1:
         Begin
            SetImageIndex(xNode, cnImgConstant);
         End;
   End;

   For i := 0 To xi.Count - 1 Do
     FindAllConstants(xi[i], xNode);
End;

Procedure TmscrRTTILibrariesTree.FindAllTypes(xi: TmscrDATAItem; Root: TTreeNode);
Var
   i: integer;
   xNode: TTreeNode;
   s: String;
Begin
   s := Copy(xi.Prop['text'], 1, 255);
   xNode := FTreeView.Items.AddChild(Root, s);
   xNode.Data := xi;

   If Root = Nil Then
      xNode.Text := xi.Name + ' (Total: ' + IntToStr(xi.Count) + ')';

   Case xNode.Level Of
      0:
         Begin
            If xi.Count = 0 Then
               SetImageIndex(xNode, cnImgBookGray)
            Else
               SetImageIndex(xNode, cnImgBook);
         End;
      1:
         Begin
            SetImageIndex(xNode, cnImgType);
         End;
   End;

   For i := 0 To xi.Count - 1 Do
      FindAllTypes(xi[i], xNode);
End;

Procedure TmscrRTTILibrariesTree.FindAllVariables(xi: TmscrDATAItem; Root: TTreeNode);
Var
   i: integer;
   xNode: TTreeNode;
   s: String;
Begin
   s := xi.Prop['text'];
   xNode := FTreeView.Items.AddChild(Root, s);
   xNode.Data := xi;

   If Root = Nil Then
      xNode.Text := xi.Name + ' (Total: ' + IntToStr(xi.Count) + ')';

   Case xNode.Level Of
      0:
         Begin
            If xi.Count = 0 Then
               SetImageIndex(xNode, cnImgBookGray)
            Else
               SetImageIndex(xNode, cnImgBook);
         End;
      1:
         Begin
            SetImageIndex(xNode, cnImgVariable);
         End;
   End;

   For i := 0 To xi.Count - 1 Do
      FindAllVariables(xi[i], xNode);
End;

Procedure TmscrRTTILibrariesTree.TreeViewExpandNodes(level: integer);
Var
   i: integer;
Begin
   If level > 1 Then
      For i := 0 To FTreeView.Items.Count - 1 Do
         If FTreeView.Items[i].Level < Level - 1 Then
            FTreeView.Items[i].Expand(False);
End;

Procedure TmscrRTTILibrariesTree.TreeViewBuild;
Begin
   FUpdating := True;
   FDATAList.Root.Clear;

   GenerateDATAContents(MSCR_GlobalScripter, FDATAList.Root);

   If FScript <> Nil Then
      GenerateDATAContents(FScript, FDATAList.Root);

   FTreeView.Items.BeginUpdate;
   FTreeView.Items.Clear;

   If ShowClasses Then
      FindAllClasses(FDATAList.Root.FindItem('Classes'), Nil);

   If ShowRecords Then
      FindAllRecords(FDATAList.Root.FindItem('Records'), Nil);

   If ShowFunctions Then
      FindAllFuncProc(FDATAList.Root.FindItem('Functions'), Nil);
                                              
   If ShowConstants Then
      FindAllConstants(FDATAList.Root.FindItem('Constants'), Nil);

   If ShowTypes Then
      FindAllTypes(FDATAList.Root.FindItem('Types'), Nil);

   If ShowVariables Then
      FindAllVariables(FDATAList.Root.FindItem('Variables'), Nil);

   If Expanded Then
      TreeViewExpandNodes(ExpandLevel);

   FTreeView.TopItem := FTreeView.Items[0];
   FTreeView.Items.EndUpdate;
   FUpdating := False;

End;

//======================================================================
//======================================================================
//======================================================================

Var
   VarLastNode: TTreeNode;
   VarLastDATAItem: TmscrDATAItem;

Procedure TmscrRTTILibrariesTree.DoOnTreeViewClick(Sender: TObject);
Var
   xItem: TmscrDATAItem;
   ss: String;
Begin
   If FDualPanel = False Then Exit;
   If Assigned(ScriptEditor) = False Then Exit;

   VarLastNode := Nil;
   FTreeView2.Items.BeginUpdate;
   FTreeView2.Items.Clear;

   If FTreeView.Selected.Data <> Nil Then
   Begin
      xItem := TmscrDATAItem(FTreeView.Selected.Data);

      If FTreeView.Selected.Parent <> Nil Then
      Begin
         ss := FTreeView.Selected.Parent.Text;

         If Pos('Classes (', ss) > 0 Then
            FindAllClasses2(xItem, Nil);

         If Pos('Records (', ss) > 0 Then
            FindAllRecords2(xItem, Nil);

         If Pos('Constants (', ss) > 0 Then
            FindAllConstants2(xItem, Nil);
                                         
         If Pos('Types (', ss) > 0 Then
            FindAllTypes2(xItem, Nil);

         If Pos('Variables (', ss) > 0 Then
            FindAllVariables2(xItem, Nil);

         If FTreeView.Selected.Parent.Parent <> Nil Then
         Begin
            ss := FTreeView.Selected.Parent.Parent.Text;
            If Pos('Functions (', ss) > 0 Then
               FindAllFuncProc2(xItem, Nil);
         End;

      End;

   End;

   FTreeView2.Items.EndUpdate;

   If VarLastNode <> Nil Then
      VarLastNode.Expand(True);
End;

Procedure TmscrRTTILibrariesTree.FindAllRecords2(xi: TmscrDATAItem; Root: TTreeNode);
Var
   i: integer;
   xNode: TTreeNode;
   s: String;
Begin
   s := xi.Prop['text'];
   xNode := FTreeView2.Items.AddChild(Root, s);

   If xNode.Level = 0 Then
      VarLastNode := xNode;

   Case xNode.Level Of
      0:
         Begin
            If xi.Count = 0 Then
               SetImageIndex(xNode, cnImgBookGray)
            Else
               SetImageIndex(xNode, cnImgBook);
         End;
      Else
         SetImageIndex(xNode, cnImgElement);
   End;

   For i := 0 To xi.Count - 1 Do
      FindAllRecords2(xi[i], xNode);
End;
        
Procedure TmscrRTTILibrariesTree.FindAllConstants2(xi: TmscrDATAItem; Root: TTreeNode);
Var
   i: integer;
   xNode: TTreeNode;
   s: String;
Begin
   s := Copy(xi.Prop['text'], 1, 255);
   xNode := FTreeView2.Items.AddChild(Root, s);

   If xNode.Level = 0 Then
      VarLastNode := xNode;

   SetImageIndex(xNode, cnImgConstant);

   For i := 0 To xi.Count - 1 Do
      FindAllTypes2(xi[i], xNode);
End;

Procedure TmscrRTTILibrariesTree.FindAllTypes2(xi: TmscrDATAItem; Root: TTreeNode);
Var
   i: integer;
   xNode: TTreeNode;
   s: String;
Begin
   s := Copy(xi.Prop['text'], 1, 255);
   xNode := FTreeView2.Items.AddChild(Root, s);

   If xNode.Level = 0 Then
      VarLastNode := xNode;

   SetImageIndex(xNode, cnImgType);

   For i := 0 To xi.Count - 1 Do
      FindAllTypes2(xi[i], xNode);
End;

Procedure TmscrRTTILibrariesTree.FindAllFuncProc2(xi: TmscrDATAItem; Root: TTreeNode);
Var
   i: integer;
   xNode: TTreeNode;
   s: String;
Begin
   s := xi.Prop['text'];
   xNode := FTreeView2.Items.AddChild(Root, s);

   If xNode.Level = 0 Then
      VarLastNode := xNode;

   If (pos('):', xNode.Text) > 0) Or ((pos('(', xNode.Text) < 1) And (pos(':', xNode.Text) > 0)) Then
      SetImageIndex(xNode, cnImgFunction)
   Else
      SetImageIndex(xNode, cnImgProcedure);

   For i := 0 To xi.Count - 1 Do
      FindAllFuncProc2(xi[i], xNode);
End;

Procedure TmscrRTTILibrariesTree.FindAllVariables2(xi: TmscrDATAItem; Root: TTreeNode);
Var
   ii, p: integer;
   s, s2: String;
   rxi: TmscrDATAItem;
   pxi: TmscrDATAItem;
Begin
   s := xi.Prop['text'];
   If s = '' Then Exit;

   //... Get TClass name .........
   p := Pos(':', s);
   If p < 1 Then Exit;
   Delete(s, 1, p);
   Delete(s, 1, 1);
   If s = '' Then Exit;

   //... Find Class...............
   rxi := FDATAList.Root.FindItem('Classes');

   pxi := Nil;
   s := s + ' =';

   For ii := 0 To rxi.Count - 1 Do
   Begin
      s2 := rxi[ii].Prop['text'];

      p := Pos(s, s2);
      If (p > 0) Then
      Begin
         pxi := rxi[ii];
         Break;
      End;

   End;

   //....Draw Class ...........
   If pxi = Nil Then Exit;
   FindAllClasses2(pxi, Nil);
End;

//=======================================================================
//=======================================================================
//=======================================================================
//=======================================================================

Function ExtractName(Var Buf: Pointer): shortstring;
Begin
   Result := '';
   If Buf = Nil Then Exit;

   SetLength(Result, byte(Buf^));
   Inc(Buf, 1);
   If Length(Result) > 0 Then
   Begin
      Move(Buf^, Result[1], Length(Result));
      Inc(Buf, Length(Result));
   End;
End;

Function ExtractParam(Var Buf: Pointer): String;
Var
   ParamType: Record
      Flags: TParamFlags;
      ParamName: shortstring;
      TypeName: shortstring;
   End;
Begin
   If Buf = Nil Then Exit;

   ParamType.Flags := TParamFlags(Pointer(Buf)^);
   Inc(Buf, SizeOf(TParamFlags));
   ParamType.ParamName := ExtractName(Buf);
   ParamType.TypeName := ExtractName(Buf);
   Result := '';
   If pfVar In ParamType.Flags Then
      Result := 'var '
   Else If pfConst In ParamType.Flags Then
      Result := 'const '
   Else If pfOut In ParamType.Flags Then
      Result := 'out '
   Else If pfArray In ParamType.Flags Then
      Result := ' array of ';
   Result := Result + ParamType.ParamName + ': ' + ParamType.TypeName;
End;

Procedure WritePropValueToStorage(Sender: TClass; PropInfo: PPropInfo; aStorage: TStrings);
Const
   CMethodKind: Array[mkProcedure..mkFunction] Of String = ('procedure ', 'function ');
Var
   TypeData: PTypeData;
   S, SS: String;
   i, ParamCount: integer;
   Buf: Pointer;
Begin
   If aStorage = Nil Then Exit;
   If Not Assigned(PropInfo^.GetProc) Then Exit;
   Case PropInfo^.PropType^.Kind Of
      tkMethod:
         Begin
            TypeData := GetTypeData(PropInfo^.PropType);
            If TypeData^.MethodKind In [Low(CMethodKind)..High(CMethodKind)] Then
            Begin
               S := CMethodKind[TypeData^.MethodKind];
               S := S + PropInfo^.Name;
               Buf := @TypeData^.ParamList;
               ParamCount := TypeData^.ParamCount;

               If ParamCount > 0 Then
               Begin
                  S := S + '(';

                  For i := 1 To ParamCount - 1 Do
                  Begin
                     SS := ExtractParam(Buf);
                     If Pos('$self', SS) < 1 Then
                        S := S + SS + '; ';
                  End;

                  S := S + ExtractParam(Buf) + ')';

                  If TypeData^.MethodKind = mkFunction Then
                     S := S + ': ' + ExtractName(Buf);
                  S := S + ';';
               End;
               aStorage.Append('event ' + PropInfo^.Name + ' := ' + S);
            End;
         End;
   End;
End;

Procedure TmscrRTTILibrariesTree.FindAllClasses2(xi: TmscrDATAItem; Root: TTreeNode);
Var
   j: integer;
   xList: TStringList;
   rxi: TmscrDATAItem;
   xxNode: TTreeNode;

   //....................................................
   Procedure xFindParentList(axi: TmscrDATAItem);
   Var
      ii, p, l: integer;
      s, s2: String;
      pxi: TmscrDATAItem;
   Begin
      s := axi.Prop['text'];
      If s = '' Then Exit;
      xList.AddObject(s, axi);

      //... Parent .........
      p := Pos('(', s);
      If p < 1 Then Exit;
      l := length(s);
      Delete(s, 1, p);
      l := length(s);
      Delete(s, l, 1);

      If s = '' Then Exit;

      pxi := Nil;
      s := s + ' =';

      For ii := 0 To rxi.Count - 1 Do
      Begin
         s2 := rxi[ii].Prop['text'];

         p := Pos(s, s2);
         If (p > 0) Then
         Begin
            pxi := rxi[ii];
            Break;
         End;

      End;

      If pxi = Nil Then Exit;
      xFindParentList(pxi);
   End;

   //....................................................
   Procedure xDrawClass(axi: TmscrDATAItem; Root: TTreeNode);
   Var
      i: integer;
      xNode: TTreeNode;
      s: String;
   Begin
      s := axi.Prop['text'];
      xNode := FTreeView2.Items.AddChild(Root, s);

      If Root = Nil Then
         VarLastNode := xNode;

      If xNode.Level = 0 Then
         VarLastDATAItem := axi;

      If axi.Count = 0 Then
         xNode.Data := axi;

      Case xNode.Level Of
         0:
            Begin
               If axi.Count = 0 Then
                  SetImageIndex(xNode, cnImgBookGray)
               Else
                  SetImageIndex(xNode, cnImgBook);
            End;
         1:
            Begin
               If pos('Constructor', xNode.Text) > 0 Then
                  SetImageIndex(xNode, cnImgConstructor) 
               Else If pos('event', xNode.Text) > 0 Then
                  SetImageIndex(xNode, cnImgEvent)
               Else If pos('procedure', xNode.Text) > 0 Then
                  SetImageIndex(xNode, cnImgProcedure)
               Else If pos('function', xNode.Text) > 0 Then
                  SetImageIndex(xNode, cnImgFunction)
               Else
                  SetImageIndex(xNode, cnImgProperty);
            End;
      End;

      //... Find Procedures/functions/Properties ............
      For i := 0 To axi.Count - 1 Do
         xDrawClass(axi[i], xNode);

   End;

   //....................................................
Begin
   If xi = Nil Then Exit;
   If FScript = Nil Then Exit;

   Try
      xList := TStringList.Create;

      rxi := FDATAList.Root.FindItem('Classes');

      xFindParentList(xi);

      If xList.Count > 0 Then
      Begin
         For j := xList.Count - 1 Downto 0 Do
         Begin
            xDrawClass(TmscrDATAItem(xList.Objects[j]), Nil);
         End;
      End;

   Finally
      xList.Free;
   End;

End;

End.
